/**
 * Super Mario World music ripping tool.
 */

#include "smwmusicrip.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define APPNAME         "SMW Music Ripper"
#define VERSION         "(beta)" // always wip :P
#define APPSHORTNAME    "smwmusicrip"
#define AUTHOR          "gocha"
#define WEBSITE         "http://gocha.s151.xrea.com/"

/*-----------------------------------------------------------*/

/**
 * search hex sequence from buffer, then returns its position.
 * \x5c (\) is a escape sequence. use \x5c\x5c for byte \x5c.
 * \x2e (.) matches with all characters.
 * \x00 means the end of the pattern.
 */
int indexOfHexPat (const byte *buf, const byte *pat, size_t bufSize)
{
  size_t patLen;
  size_t index;
  size_t bufIndex;
  size_t patIndex;
  byte bufB, patB;
  bool matchFailed;

  // calc pattern size first
  patLen = 0;
  index = 0;
  while (pat[index] != '\0') {
    switch (pat[index]) {
    case 0x5c:
      index += 2;
      break;
    default:
      index++;
    }
    patLen++;
  }

  // failure: pattern is larger than buffer
  if (bufSize < patLen)
    return -1;

  // search the pattern from beginning
  for (index = 0; index <= bufSize - patLen; index++) {
    patIndex = 0;
    matchFailed = false;
    for (bufIndex = index; !matchFailed && bufIndex < index + patLen; bufIndex++) {
      bufB = buf[bufIndex];
      patB = pat[patIndex];
      switch (patB) {
      case 0x2e:
        break;
      case 0x5c:
        patB = pat[++patIndex];
        // fall through
      default:
        if (bufB != patB)
          matchFailed = true;
      }
      patIndex++;
    }
    if (!matchFailed)
      return (int) index;
  }
  // not found
  return -1;
}

//----

static int spcSongIndexMaster = 0;  // 0 = read from port
static int spcSongFrom = 0xf6;      // (port2 = BGM)
static int spcSongList = 0x135e;    // song table
static int spcAddrAdjust = 0;       // offset (0 for no fix)
static int spcVcmdLenTbl = 0x0fc2;  // vcmd length table (for readahead)

static byte spcPitchSlide = 0xdd;
static byte spcCallCmd = 0xe9;

static byte aram[0x10000];          // ARAM output
static const byte ARAM[0x10000];    // SPC ARAM (read only)
static byte spcInitByte = 0xff;

static int spcSongIndex;
static int spcSongEntry;
static int spcBlockAddr;
static int spcSeqAddr[8];
static int spcSeqDur[8];
static int spcSeqTick[8];
static int spcSeqSubAddr[8];
static int spcSeqSubLoop[8];
static int spcSeqRetAddr[8];

/** ref ARAM n bytes. */
const byte *refARAMBytes (int index, int size)
{
    memcpy(&aram[index], &ARAM[index], size);
    return &ARAM[index];
}

/** ref a byte from ARAM. */
int refARAMByte (int index)
{
    aram[index] = ARAM[index];
    return ARAM[index];
}

/** ref a word from ARAM. */
int refARAMWord (int index)
{
    memcpy(&aram[index], &ARAM[index], 2);
    return mget2l(&ARAM[index]);
}

/** ref an addr from ARAM. */
int refARAMWordAddr (int index)
{
    int addr = mget2l(&ARAM[index]);
    mput2l(addr ? ((addr + spcAddrAdjust) & 0xffff) : 0, &aram[index]);
    return addr;
}

/** read block. */
bool smwReadBlock (int addr)
{
    int ch;
    bool nullBlock = true;

    for (ch = 0; ch < 8; ch++) {
        spcSeqAddr[ch] = refARAMWordAddr(addr + ch * 2);
        if (spcSeqAddr[ch])
            nullBlock = false;
    }
    return !nullBlock;
}

/** goto next block. */
bool smwNextBlock (void)
{
    int loopCount = 0;
    int entryAddr;
    int ch;

    for (ch = 0; ch < 8; ch++) {
        spcSeqTick[ch] = 0;
        spcSeqSubLoop[ch] = 0;
    }

    entryAddr = refARAMWord(spcBlockAddr);
    if ((entryAddr & 0xff00) == 0) {
        loopCount = utos1(entryAddr);
        if (loopCount == 0) {
            return false;
        }
        spcBlockAddr += 2;
        entryAddr = refARAMWord(spcBlockAddr);
    }
    refARAMWordAddr(spcBlockAddr); // adjust addr
    spcBlockAddr += 2;

    if (loopCount < 0) {
        // consider it loop-point
        return false;
    }
    return smwReadBlock(entryAddr);
}

/** parse sequence. */
void smwDoBlock (void)
{
    int ch;
    bool endBlock = false;
    byte vcmd, c;
    bool readahead = false;

    while (!endBlock) {
        for (ch = 0; !endBlock && ch < 8; ch++) {
          if (spcSeqAddr[ch]) do {
            if (!readahead && spcSeqTick[ch] > 0) {
              spcSeqTick[ch]--;
              break;
            }

            vcmd = refARAMByte(spcSeqAddr[ch]);
            spcSeqAddr[ch]++;

            if (vcmd == 0x00) {
                if (spcSeqRetAddr[ch] == 0) {
                    if (!spcSeqTick[ch]) {
                        endBlock = true;
                    }
                    else {
                        // stop readahead at this point
                        spcSeqAddr[ch]--;
                        readahead = false;
                    }
                    break;
                }
                else {
                    if (spcSeqSubLoop[ch])
                        spcSeqSubLoop[ch]--;
                    spcSeqAddr[ch] = spcSeqSubLoop[ch] ? spcSeqSubAddr[ch] : spcSeqRetAddr[ch];
                }
            }
            else if (vcmd >= 0x01 && vcmd < 0x80) {
                // length
                spcSeqDur[ch] = vcmd;

                c = ARAM[spcSeqAddr[ch]];
                if (c < 0x80) {
                    // skip dur/vel
                    refARAMByte(spcSeqAddr[ch]); // mark
                    spcSeqAddr[ch]++;
                }
            }
            else if (vcmd >= 0x80 && vcmd < 0xda) {
                // note
                spcSeqTick[ch] = spcSeqDur[ch];
            }
            else if (vcmd == spcPitchSlide) {
                refARAMBytes(spcSeqAddr[ch], 3);
                spcSeqAddr[ch] += 3;
            }
            else if (vcmd == spcCallCmd) {
                spcSeqSubAddr[ch] = refARAMWordAddr(spcSeqAddr[ch]);
                spcSeqSubLoop[ch] = refARAMByte(spcSeqAddr[ch] + 2);
                fprintf(stderr, "Call: at $%04x to $%04x for %d times\n", spcSeqAddr[ch] - 1, spcSeqSubAddr[ch], spcSeqSubLoop[ch]);
                spcSeqAddr[ch] += 3;
                spcSeqRetAddr[ch] = spcSeqAddr[ch];
                spcSeqAddr[ch] = spcSeqSubAddr[ch];
            }
            else {
                // readahead
                int vcmdLen = ARAM[spcVcmdLenTbl + (vcmd - 0xda)] - 1;
                refARAMBytes(spcSeqAddr[ch], vcmdLen);
                spcSeqAddr[ch] += vcmdLen;
            }

            vcmd = ARAM[spcSeqAddr[ch]];
            readahead = (vcmd == 0xdd) || (vcmd == 0x00);
          } while (!spcSeqTick[ch] || readahead);
        }
    }
}

/** rip smw music from ARAM. */
bool smwRipMusicMain (void)
{
    int ch;

    memset(aram, spcInitByte, 0x10000);
    spcSongIndex = spcSongIndexMaster;
    if (spcSongIndex == 0) {
        spcSongIndex = ARAM[spcSongFrom];
        fprintf(stderr, "Autodetect: song index $%02x\n", spcSongIndex);
    }
    spcSongEntry = mget2l(&ARAM[spcSongList + spcSongIndex * 2]);

    for (ch = 0; ch < 8; ch++) {
        spcSeqDur[ch] = 1;
        spcSeqRetAddr[ch] = 0;
    }

    spcBlockAddr = spcSongEntry;
    while (smwNextBlock()) {
        smwDoBlock();
    }

    return true;
}

/*-----------------------------------------------------------*/

static int gArgc;
static char **gArgv;
static bool manDisplayed = false;

typedef bool (*CmdDispatcher) (void);

typedef struct TagCmdOptDefs {
    char *name;
    char shortName;
    int numArgs;
    CmdDispatcher dispatch;
    char *syntax;
    char *description;
} CmdOptDefs;

static bool cmdOptHelp (void);
static bool cmdOptSong (void);
static bool cmdOptPort0 (void);
static bool cmdOptPort2 (void);
static bool cmdOptOffset (void);

static CmdOptDefs optDef[] = {
    { "help", '\0', 0, cmdOptHelp, "", "show usage" },
    { "song", '\0', 1, cmdOptSong, "<index>", "specify target song" },
    { "port0", '0', 1, cmdOptPort0, "", NULL /*"read song index from port0"*/ },
    { "port2", '2', 1, cmdOptPort2, "", NULL /*"read song index from port2 (default)"*/ },
    { "offset", '\0', 1, cmdOptOffset, "<addr>", "adjust jump addresses (hex)" },
};

//----

/** display how to use. */
void man (void)
{
    const char *cmdname = APPSHORTNAME;
    int op;

    fprintf(stderr, "\n");
    fprintf(stderr, "%s - %s %s\n", APPSHORTNAME, APPNAME, VERSION);
    fprintf(stderr, "Syntax: %s (options) [spc input] [ARAM output]\n", cmdname);
    fprintf(stderr, "%s\n", WEBSITE);

    fprintf(stderr, "\n");
    fprintf(stderr, "Options:\n");
    for (op = 0; op < countof(optDef); op++) {
        if (optDef[op].description) {
            if (optDef[op].dispatch) {
                fprintf(stderr, " %s%c  %s%-8s  %-15s  %s\n",
                    (optDef[op].shortName != '\0') ? "-" : " ",
                    (optDef[op].shortName != '\0') ? optDef[op].shortName : ' ',
                    optDef[op].name ? "--" : "  ",
                    optDef[op].name ? optDef[op].name : "",
                    optDef[op].syntax ? optDef[op].syntax : "",
                    optDef[op].description ? optDef[op].description : "");
            }
            else
                fprintf(stderr, "\n");
        }
    }

    manDisplayed = true;
}

/** display about application. */
void about (void)
{
    const char *cmdname = APPSHORTNAME;

    fprintf(stderr, "%s - %s %s\n", APPSHORTNAME, APPNAME, VERSION);
    fprintf(stderr, "Programmed by %s - %s\n", AUTHOR, WEBSITE);
    fprintf(stderr, "Syntax: %s (options) [spcfile] [midfile] (htmlfile)\n", cmdname);
}

//----

/** show usage. */
static bool cmdOptHelp (void)
{
    man();
    return true;
}

/** handle command-line options. */
static bool handleCmdLineOpts (void)
{
    int op;

    // dispatch options
    while (gArgc > 0 && gArgv[0][0] == '-') {
        bool shortOpt = (gArgv[0][1] != '-');
        int chIndex;

        // match for each option
        for (chIndex = 1; chIndex < (int) (shortOpt ? strlen(gArgv[0]) : 2); chIndex++) {
            bool unknown = true;

            for (op = 0; op < countof(optDef); op++) {
                if (optDef[op].dispatch
                        && ((!shortOpt && optDef[op].name && strcmp(&gArgv[0][2], optDef[op].name) == 0)
                        || (shortOpt && optDef[op].shortName != '\0' && gArgv[0][chIndex] == optDef[op].shortName))) {
                    unknown = false;
                    if (!shortOpt) {
                        gArgc--;
                        gArgv++;
                        if (gArgc >= optDef[op].numArgs) {
                            if (!optDef[op].dispatch())
                                return false;
                            gArgc -= optDef[op].numArgs;
                            gArgv += optDef[op].numArgs;
                        }
                        else {
                            fprintf(stderr, "Error: too few arguments for option \"--%s\".\n", optDef[op].name);
                            gArgv += gArgc;
                            gArgc = 0;
                            return false;
                        }
                    }
                    else {
                        assert(optDef[op].numArgs == 0);
                        if (!optDef[op].dispatch())
                            return false;
                    }
                    break;
                }
            }
            if (unknown) {
                if (!shortOpt)
                    fprintf(stderr, "Error: unknown option \"%s\".\n", gArgv[0]);
                else
                    fprintf(stderr, "Error: unknown option \"-%c\".\n", gArgv[0][chIndex]);
                gArgc--;
                gArgv++;
                return false;
            }
        }
        if (shortOpt) {
            gArgc--;
            gArgv++;
        }
    }
    return true;
}

/** specify song index. */
static bool cmdOptSong (void)
{
    spcSongIndexMaster = strtol(gArgv[0], NULL, 0);
    return true;
}

/** set to read song index from port0. */
static bool cmdOptPort0 (void)
{
    spcSongFrom = 0xf4;
    //spcSongList = 0x135e;
    return true;
}

/** set to read song index from port2. */
static bool cmdOptPort2 (void)
{
    spcSongFrom = 0xf6;
    spcSongList = 0x135e;
    return true;
}

/** set adjust amount for jump addresses. */
static bool cmdOptOffset (void)
{
    spcAddrAdjust = strtol(gArgv[0], NULL, 16);
    return true;
}

//----

/** application main. */
int main (int argc, char *argv[])
{
    bool result;
    FILE *spcFile;
    FILE *aramFile;
    byte spcHead[0x100];

    // handle options
    gArgc = argc - 1;
    gArgv = argv + 1;
    result = handleCmdLineOpts();

    // too few or much args
    if (gArgc != 2 || !result) {
        if (!manDisplayed) {
            about();
            fprintf(stderr, "Run with --help, for more details.\n");
            return (argc == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
        }
        else
            return EXIT_SUCCESS;
    }

    fprintf(stderr, "%s:\n", gArgv[0]);
    spcFile = fopen(gArgv[0], "rb");
    if (!spcFile) {
        fprintf(stderr, "Error: spc file couldn't be opened\n");
        return EXIT_FAILURE;
    }
    fread(spcHead, 0x100, 1, spcFile);
    if (memcmp(spcHead, "SNES-SPC700 Sound File Data", 27) != 0) {
        fprintf(stderr, "Error: spc input is invalid\n");
        fclose(spcFile);
        return EXIT_FAILURE;
    }
    fseek(spcFile, 0x100, SEEK_SET);
    fread((byte *) ARAM, 0x10000, 1, spcFile);

    result = smwRipMusicMain();
    if (result) {
        aramFile = fopen(gArgv[1], "wb");
        if (!aramFile) {
            fprintf(stderr, "Error: ARAM file couldn't be opened\n");
            return EXIT_FAILURE;
        }

        fwrite(aram, 0x10000, 1, aramFile);
        fclose(aramFile);
    }

    fclose(spcFile);
    return result ? EXIT_SUCCESS : EXIT_FAILURE;
}
